//
//  Copyright (C) 2012 Vitor Rolla
//  vitorgr [at] dei.uc.pt
//  University of Coimbra - Portugal
//


#ifndef NETWORKLAYER_MOFO_MOFO_H
#define NETWORKLAYER_MOFO_MOFO_H


#include <omnetpp.h>
#include <algorithm>

#include "IPAddress.h"
#include "IP.h"
#include "IInterfaceTable.h"
#include "InterfaceTableAccess.h"
#include "Ieee802Ctrl_m.h"
#include "ICMPMessage.h"

#include "networklayer/mofo/MOFO_MessageBlock.h"

#include "networklayer/mofo/MOFO_CTRL_m.h"
#include "networklayer/mofo/MOFO_RQST_m.h"
#include "networklayer/mofo/MOFO_Timeout_m.h"
#include "networklayer/mofo/MOFO_Timer.h"
#include "networklayer/mofo/MOFO_Packet_m.h"
#include "networklayer/mofo/MOFO_ACK_m.h"
#include "networklayer/mofo/AppPkt_m.h"


using namespace MOFO;


namespace MOFO {

//===========================================================================================
// class QueueElement: defines a data type for all queued data messages
//===========================================================================================
class QueueElement : public cPacket
{
	public:
		QueueElement(){ msgId = 0; msgCreationTime = 0; destAddr = 0; srcAddr = 0; nOfHops = 0; initNumberOfCopies = 0; numberOfCopies = 0; obj = NULL; };

		QueueElement(unsigned int msgId, simtime_t msgCreationTime, unsigned int destAddr, unsigned int srcAddr, unsigned int nOfHops, unsigned int initNumberOfCopies, unsigned int numberOfCopies, MOFO_Packet* obj){
			this->msgId = msgId;
			this->msgCreationTime = msgCreationTime;
			this->destAddr = destAddr;
			this->srcAddr = srcAddr;
			this->numberOfCopies = numberOfCopies;
			this->nOfHops = nOfHops;
			this->initNumberOfCopies = initNumberOfCopies;
			this->obj = obj;
		};
		virtual ~QueueElement(){};
		unsigned int getMsgId(){ return msgId; };
		simtime_t getCreationTime(){ return msgCreationTime; };
		unsigned int getDestAddr(){ return destAddr; };
		unsigned int getSrcAddr(){ return srcAddr; };
		unsigned int getNumberOfCopies(){ return numberOfCopies; };
		unsigned int getNOfHops(){ return nOfHops; };
		unsigned int getInitNumberOfCopies(){ return initNumberOfCopies; };
		MOFO_Packet* getObject(){ return obj; };
		void setNumberOfCopies(){ this->numberOfCopies = this->numberOfCopies/2; this->obj->setNOfCopies(this->numberOfCopies);}; //Switch: Two-hop vs. SnW

	private:
		unsigned int msgId;
		simtime_t msgCreationTime;
		unsigned int destAddr;
		unsigned int srcAddr;
		unsigned int numberOfCopies;
		unsigned int nOfHops;
		unsigned int initNumberOfCopies;
		MOFO_Packet* obj;

};

}


class MOFO_Net : public cSimpleModule
{
  public:
    int numInitStages() const  {return 4;}
    void initialize(int);
    void finish();


    /** @brief Function called whenever a message arrives at the module */
    void handleMessage(cMessage * msg);

    /** @brief set the address in MOFO network */
    void setMyAddr(unsigned int myAddr);

    /** @brief guesses which router the given address belongs to, might return 0 */
	cModule* getRouterByAddress(IPAddress address);

  private:

      /** @brief Handle self messages such as timer... */
      virtual void handleSelfMsg(cMessage*);



      struct compareEntries {
          bool operator ()(MOFO_MessageBlock lhs, MOFO_MessageBlock rhs) { return (lhs.getNOfCopies() > rhs.getNOfCopies()); }
      };



      /** @brief Handle messages from upper and lower layers */
      virtual void handleUpperMsg(cPacket*);
      virtual void handleLowerMsg(cPacket*);

      /** @brief Function handles CTRL msgs */
      virtual void handleLowerCTRL(MOFO_CTRL*);
      /** @brief Function handles CTRL msgs */
	  virtual void handleLowerRQST(MOFO_RQST*);
      /** @brief Function handles DTN Data msgs */
      virtual void handleLowerPacket(MOFO_Packet*);
      virtual void handleLowerACK(MOFO_ACK *);

      //Insert DTN messages from Upper layer data queue.
      void insertMsg(MOFO_Packet *packet);

      /** @brief Function encapsulates DTN data msgs from upper layer */
      MOFO_Packet* encapsMsg(cPacket*);

      /** @brief Function sends DTN data msgs to upper layer */
      void sendUp(cPacket*);

      /** @brief Function sends messages to lower layer (wlan layer) */
      void sendDown(cPacket*, int);

      /** @brief generates and sends down a new CTRL msg */
      void sendCTRL();

      //Check and send Queued Packets to Original CTRL msg node.
      void checkAndSendQueuedPktsToDest(unsigned int destinationAddress);


      //===============================================================================
      // MEMBER VARIABLES
      //===============================================================================

      //@when CTRL msg period expired
      MOFO_Timer* ctrlTimeout;

      /** @brief address of the node */
      unsigned int myAddr;

      /** @Number of queued msgs */
      int numQueuedMsgs;

      /** @brief vector contains the queued data packets, waiting */
      cQueue * queuedElements;

      /** @brief defines the message block size */
      unsigned int mblockSize;

      /** @Collect statistics */
      int totalPacketsSent; /**< total number of packets sent down to lower layer */
      int totalPacketsRcvd; /**< total number of packets sent up to app layer */
      int totalPacketsDeleted; /**< total number of packets deleted from buffer*/
      int totalPacketsReplicated; /**< total number of packets replicated */
      int totalBytesSent; /**< total number of bytes sent down to lower layer */
      int totalSelfishPacketsDelivery;
      int selfishIsoDelivery;
      int unselfishIsoDelivery;

      double totalRcvdReward;

      int statsCTRLSent; /**< number of generated MOFO CTRL msg */
      int statsCTRLRcvd; /**< number of consumed MOFO CTRL msg */
      int statsRQSTSent; /**< number of generated MOFO RQST msg */
	  int statsRQSTRcvd; /**< number of consumed MOFO RQST msg */
      int statsTrafficSent; /**< number of injected Application-layer messages */
      int statsTrafficRcvd; /**< number of received Application-layer messages (for this node) */
      int statsACKSent; /**< number of generated MOFO CTRL msg */
      int statsACKRcvd; /**< number of consumed MOFO CTRL msg */

      cOutVector bufferValueVec;
      cOutVector numOfNonSelfishPacketsOnBuffer;
      simtime_t dataLatencySelfish;
      simtime_t dataLatencyUnselfish;
      int dataSamplesSelfish;;
      int dataSamplesUnselfish;
      cOutVector dataDelayVec;
      cOutVector nOfHopsVector;


      /** @Message Control*/

      typedef std::vector<unsigned int> MessageVector;
      MessageVector messageVector;

      /** @NED parameters*/
      int RESPONSIBLE_ADDRESSES_PREFIX; /**< NED parameter: netmask of network this MOFO router is responsible for, -1 for self only */
      const char* MOFO_INTERFACES; /**< NED parameter: list of interfaces to run MOFO on, separated by a single space character */
      IPAddress AUTOASSIGN_ADDRESS_BASE; /**< NED parameter: start of address range from which to automatically assign an address to the MOFO_INTERFACES */
      int CTRL_MSG_PERIOD; /**< NED parameter: CTRL msg frequency  */
      int BUFFER_SIZE_PACKETS; /**< NED configuration parameter: maximum number of queued packets, -1 for no limit */
      int BUFFER_SIZE_BYTES; /**< NED configuration parameter: maximum total size of queued packets, -1 for no limit */
      int NUM_OF_COPIES; /**< Number of Copies of on msg allowed in the system  */
      double LAMBDA;
      /** @brief gate id*/
		//@{
		int mUppergateIn;
		int mUppergateOut;
		int mLowergateIn[2];
		int mLowergateOut[2];
		//@}

  };

  #endif

