//
//  Copyright (C) 2012 Vitor Rolla
//  vitorgr [at] dei.uc.pt
//  University of Coimbra - Portugal
//

#ifndef MOFO_Timer_H
#define MOFO_Timer_H

#include <iostream>
#include <omnetpp.h>
#include "MOFO_Timeout_m.h"

namespace MOFO{

//===========================================================================================
// simple, OMNeT++-aware timer class.
// Starting a MOFO_Timer will schedule a MOFO_Timeout cMessage to be sent to the parent when
// the timer expires. Warning: This cMessage is recycled for every iteration - do not delete!
//===========================================================================================
class MOFO_Timer : public cObject
{
  public:
    /**
     * @param parent specifies the cSimpleModule that will receive a MOFO_Timeout message when this timer expires
     * @param name helps differentiate MOFO_Timer instances
     * @param interval sets the default interval after which to expire the MOFO_Timer
     */
    MOFO_Timer(cSimpleModule* parent, std::string name, simtime_t interval = 0);

    ~MOFO_Timer();

    /** @brief inherited from cObject */
    virtual const char* getFullName() const;

    /** @brief inherited from cObject */
    virtual std::string info() const;

    /** @brief inherited from cObject */
    virtual std::string detailedInfo() const;

    /** @brief returns whether the given timeout was started, but is not yet expired */
    bool isRunning();

    /** @brief returns whether the given timeout is to be considered expired */
    bool isExpired();

    /** @brief (re-)starts the timeout */
    void start(simtime_t interval = 0);

    /** @brief cancels the timeout */
    void cancel();

    /** @brief checks if the given cMessage is associated with this MOFO_Timer */
    bool owns(const cMessage* message) const;

    /** @brief returns the last set interval of this timer */
    simtime_t getInterval() const;

  protected:
    cSimpleModule* parent; /**< cSimpleModule that will receive cMessage instances that indicate timeouts */
    char* name; /**< descriptive name */
    simtime_t interval; /**< last set interval after which to expire this MOFO_Timer */
    MOFO_Timeout* message; /**< cMessage to be scheduled to remind simulation core of timeout */
    simtime_t expiresAt; /**< point in simulation time from where on this MOFO_Timer will be considered expired */
    bool active; /**< false if this MOFO_Timer is not currently active */

  public:
    friend std::ostream& operator<< (std::ostream& os, const MOFO_Timer& o);

};

std::ostream& operator<< (std::ostream& os, const MOFO_Timer& o);
}
#endif

